#!/usr/bin/env perl

#   CGIForum -- light-weight discussion board
#   Copyright (C) 2000-2003, 2006 Markus Triska (triska@gmx.at)
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA

use strict;
use warnings;

use Time::Local;
use CGI;

use Fcntl ':flock';

# change these settings in "cgiforum.conf"

my $basedir	       = '../cgiforum/';
my $period_new	       = 2;
my $default_day_limit  = 30;
my $default_headline   = 'None';
my $default_author     = 'Anonymous';
my %start_entries      = ();
my $load_text	       = 0;
my $only_registered    = 0;
my $allow_registration = 1;
my $encryption_method  = 'xor12';

my @child_parent = ();
my %entry_info = ();
# the section we are currently working in
my $section = '';
my %expand_day = (
    Mon => 'Monday', Tue => 'Tuesday', Wed => 'Wednesday', Thu => 'Thursday',
    Fri => 'Friday', Sat => 'Saturday', Sun => 'Sunday' );

my %month_number = (
    Jan => '01', Feb => '02', Mar => '03', Apr => '04', May => '05',
    Jun => '06', Jul => '07', Aug => '08', Sep => '09', Oct => '10',
    Nov => '11', Dec => '12' );

my $cgi = new CGI;



sub filter_input
{
    # annihilate ".." etc. in section names, which could be used to
    # gain (read) access to upper directories

    my $string = shift;

    # contributed by b0iler@hotmail.com
    $string =~ s/([\&;\`'\|\"*\?\~\^\(\)\[\]\{\}\$\n\r])/\\$1/g;
    $string =~ s/\0//g;
    $string =~ s/\.\.//g;

    return $string;
}



sub encrypt_xor12
{
    # XOR each character with 12
    my @args = split //, shift;
    my $retval = '';
    foreach my $c (@args) { $retval .= chr(ord($c) ^ 0x12); }
    return $retval;
}


sub verify_xor12
{
    my ($given, $stored) = @_;
    return $given eq (encrypt_xor12 $stored);
}


sub encrypt_crypt
{
    # use Perl's built-in "crypt" function
    my $salt = join '', ('.', '/', 0..9, 'A'..'Z', 'a'..'z')[rand 64,rand 64];
    return crypt shift, $salt;
}

sub verify_crypt
{
    my ($given, $stored) = @_;
    return crypt($given, $stored) eq $stored;
}


# dispatching -- register new encryption/verification functions here

sub encrypt_password
{
    $encryption_method eq "crypt" and return encrypt_crypt shift;
    $encryption_method eq "xor12" and return encrypt_xor12 shift;
    print "invalid encryption method: $encryption_method" and die;
}

sub verify_password
{
    my ($given, $stored) = @_;
    $encryption_method eq "crypt" and return verify_crypt $given, $stored;
    $encryption_method eq "xor12" and return verify_xor12 $given, $stored;
    print "invalid encryption method: $encryption_method" and die;
}



sub age_in_days
{
    # age in days of currently loaded entry
    my $newstime = timelocal(0, 0, 0, $entry_info{'day'}, $entry_info{'month'} - 1, $entry_info{'year'} - 1900);
    my $ltime = localtime;
    if (substr ($ltime, 8, 1) eq ' ') {	substr ($ltime, 8, 1) = '0'; }
    my ($weekday, $monthstr, $day, $Time, $year) = split / /, $ltime;
    my $nummonth = $month_number{$monthstr};
    my $currenttime = timelocal(0, 0, 0, $day, $nummonth - 1, $year - 1900);
    my $diff = $currenttime - $newstime;
    my $seconds = $diff % 60;
    $diff = ($diff - $seconds) / 60;
    my $minutes = $diff % 60;
    $diff = ($diff - $minutes) / 60;
    my $hours = $diff % 24;
    return ($diff - $hours) / 24;
}

sub within_date_limits
{
    # true iff currently loaded entry is within date limits
    my $younger_than_start = $entry_info{'start'} == -1 ||
	age_in_days() <= $entry_info{'start'};

    return $younger_than_start && age_in_days >= $entry_info{'end'};
}

sub fill_entry_time
{
    my $time = shift;

    if (substr ($time, 8, 1) eq ' ') { substr ($time, 8, 1) = '0'; }

    my ($weekday, $monthstr, $day, $Time, $year) = split / /, $time;
    $entry_info{'weekday'} = $expand_day{substr $weekday, 0, 3};
    $entry_info{'time'} = $Time;
    $entry_info{'day'} = $day;
    $entry_info{'monthstr'} = $monthstr;
    $entry_info{'month'} = $month_number{$monthstr};
    $entry_info{'year'} = $year;
    $entry_info{'hour'} = substr $Time, 0, 2;     # 16:03:22
    $entry_info{'minute'} = substr $Time, 3, 2;
    $entry_info{'second'} = substr $Time, 6, 2;
    $entry_info{'newsdate'} = $year . $month_number{$monthstr} . $day;
}

sub load_entry_time
{
    my $entrynumber = shift;
    my $entrypath = $basedir . "sections/$section/entries/$entrynumber";
    if (-e "$entrypath.inf") {
	open ENTRYINFO, "<$entrypath.inf";
	my $infoline = <ENTRYINFO>;
	close ENTRYINFO;
	chomp $infoline;
	my ($author_, $time, $email_) = split /\|/, $infoline;
	fill_entry_time $time;
    }
}

sub load_entry
{
    my ($entrynumber, $loadtext) = @_;
    $entry_info{'entrynumber'} = $entrynumber;

    my $entrypath = $basedir . "sections/$section/entries/$entrynumber";

    if ($loadtext) { $entry_info{'text'} = file_to_scalar("$entrypath.msg"); }

    my $hdlfile = $basedir . "sections/$section";
    if ($entrynumber == 0) {
	$hdlfile .= '/section.txt';
    } else {
	$hdlfile .= "/entries/$entrynumber.hdl";
    }

    open HDL, "<$hdlfile" or print "couldn't open $hdlfile\n" and die;
    my $headline = <HDL> || $default_headline;
    close HDL;

    chomp $headline;
    $entry_info{'headline'} = $headline;

    my $infoline = '||';
    if (-e "$entrypath.inf") {
	open ENTRYINFO, "<$entrypath.inf";
	$infoline = <ENTRYINFO>;
	close ENTRYINFO;
	chomp $infoline;
    }

    my ($author, $time, $email) = split /\|/, $infoline;
    fill_entry_time $time;
    my $authornum = 0;
    if ($author =~ /^REGISTERED/) {
	($authornum = $author) =~ s/^REGISTERED//;
	chomp $authornum;
	open AUTHOR, "<$basedir". "users/$authornum.syn"
	    or print "couldn't open author file $authornum.syn\n" and die;
	$author = <AUTHOR>;
	close AUTHOR;
	chomp $author;
    }
    if ($email =~ /seeeml/) {
	open EMAIL, "<$basedir" . "users/$authornum.eml";
	$email = <EMAIL> || "";
	close EMAIL;
	chomp $email;
    }

    $author or $author = $default_author;
    $entry_info{'author'} = $author;
    $entry_info{'email'} = $email;

    my $numhits = 0;
    # prior to version 1.04, no "cnt"-file was written by default
    if (-e "$entrypath.cnt") {
	open COUNT, "<$entrypath.cnt";
	$numhits = <COUNT>;
	close COUNT;
	chomp $numhits;
    }
    $entry_info{'numhits'} = $numhits;
}


sub build_tree
{

    # entries.db contains lines of the form "child|parent". child and
    # parent are (numeric) IDs of entries. Here, we build a list of
    # "ID|level" elements with all subordinate entries.

    my ($currentthread, $level) = @_;

    my @tree = ();
    foreach my $entry (grep /\|$currentthread\n/, @child_parent) {
	my ($child, $p_) = split /\|/, $entry;
	unless (-e $basedir . "/sections/$section/entries/$child.del") {
	    # add entry if within date limits
	    load_entry_time $child;
	    if (within_date_limits) {
		push @tree, "$child\|$level";
		push @tree, build_tree($child, $level + 1);
	    } else {
		push @tree, build_tree($child, $level);
	    }
	}
    }
    return @tree;
}

sub htmlify
{
    my $string = shift;

    $string =~ s//\&auml\;/g; $string =~ s//\&Auml\;/g;
    $string =~ s//\&ouml\;/g; $string =~ s//\&Ouml\;/g;
    $string =~ s//\&uuml\;/g; $string =~ s//\&Uuml\;/g;

    $string =~ s/\n/\<BR\>/g;
    $string =~ s//\&szlig\;/g;

    # to not interpret "<" and ">" as HTML-tag delimiters:
    # $string =~ s/</\&lt\;/g;
    # $string =~ s/>/\&gt\;/g;

    # ... add more special characters here

    return $string;
}


sub subst_variables
{
    my $template = shift;
    $template =~ s/\#\#(.*?)\#\#/htmlify (exists $entry_info{$1} ? $entry_info{$1} : "Warning...could NOT substitute $1")/ge;
    return $template;
}

sub file_to_scalar
{
    my $file = pop;
    my $scalar = '';
    if (!-e $file) {
	print "<b><font color=\"\#FF0000\"> Warning - file $file was not found. Please install it. It is included in this package of CGIForum. </font></b>" and exit;
    } else {
	open FILE, "<$file";
	while (my $l = <FILE>) { $scalar .= $l; }
	close FILE;
    }
    return $scalar;
}

sub customized_html
{
    my $file = shift;
    my $custom = $basedir . "sections/$section/html/$file";
    return (-e $custom) ? $custom : $basedir . "html/$file";
}

sub print_html
{
    foreach my $file (@_) {
	my $filename = customized_html $file;
	if (-e $filename) {
	    open HTMLFILE, "<$filename"
		or print "Cannot read $filename -- please install.\n" and die;
	    while (my $l = <HTMLFILE>) { print subst_variables $l; }
	    close HTMLFILE;
	} else {
	    print "File $filename does not exist!\n" and exit;
	}
    }
    return 1;
}


sub load_template
{
    return file_to_scalar customized_html shift;
}


sub print_entries
{
    my $startthread = shift;
    my $db = $basedir . "sections/$section/entries/entries.db";
    if (-e $db && ($#child_parent == -1)) {
	# (if we are called by print_entry, the list is already trimmed)
	my $startentry = $startthread;
	if ($startthread == 0 && $entry_info{'start'} == $default_day_limit) {
	    $startentry = $start_entries{$section} || 0;
	}

	open ENTRYBASE, "<$db";
	while (<ENTRYBASE>) {
	    my ($child, $parent_) = split /\|/;
	    if ($child >= $startentry) { push @child_parent, $_; }
	}
	close ENTRYBASE;
    }

    # reverse chronological order for all children of "0" when
    # printing whole section
    if ($startthread == 0 && $#child_parent >= 0) {
	my $last = $#child_parent;
	my $i = 0;
	while ($i < $last) {
	    if ($child_parent[$i] =~ /\|0\n/)  {
		# find last child of 0
		while ($child_parent[$last] !~ /\|0\n/) { $last--; }

		# and swap them
		my $tmp = $child_parent[$i];
		$child_parent[$i] = $child_parent[$last];
		$child_parent[$last] = $tmp;
		$last--;
	    }
	    $i++;
	}
    }

    my @tree = build_tree $startthread, 1;
    if ($startthread == 0) {
	# put everything in one large "base-thread":
	# unshift @tree, "$startthread\|0";
    } else {
	unshift @tree, "$startthread\|0";
    }


    my ($parent, $parent_active, $parent_end) = map { load_template $_; }
    ('parententry.tpl', 'parententry.tpa', 'parententry.end');

    my ($normal, $normal_active) = map { load_template $_; }
    ('normalentry.tpl', 'normalentry.tpa');

    my ($parent_new, $parent_new_active, $normal_new, $normal_new_active) =
	map { load_template $_; }
    ('parententry_new.tpl', 'parententry_new.tpa', 'normalentry_new.tpl',
     'normalentry_new.tpa');


    my $prev_level = 0;
    for (my $index = 0; $index <= $#tree; $index++) {
	my ($number, $level) = split /\|/, $tree[$index];
	while ($prev_level > $level) {
	    print subst_variables $parent_end;
	    $prev_level--;
	}
	$prev_level = $level;
	my $is_parent = 0;
	my $nextentry = $tree[$index + 1];
	if ($nextentry) {
	    my ($nextnumber, $nextlevel) = split /\|/, $nextentry;
	    $is_parent = $nextlevel > $level;
	}

	load_entry $number, $load_text;

	my $active = $entry_info{'topmost'} == $number;

	print subst_variables $is_parent
	    ? ((age_in_days() <= $period_new) ?
	       ($active ? $parent_new_active : $parent_new)
	       : ($active ? $parent_active : $parent))
	    : ((age_in_days() <= $period_new) ?
	       ($active ? $normal_new_active : $normal_new)
	       : ($active ? $normal_active : $normal));
    }

    my $start_level = (split /\|/, $tree[0])[1];
    while ($prev_level > $start_level) {
	print subst_variables $parent_end;
	$prev_level--;
    }
}

sub lock { flock shift, LOCK_EX; }

sub lock_sh { flock shift, LOCK_SH; }

sub unlock { flock shift, LOCK_UN; }

sub print_entry
{
    my $entrynumber = shift;
    my $db = $basedir . "sections/$section/entries/entries.db";

    open ENTRYBASE, $db or die "cannot open entries.db";
    @child_parent = <ENTRYBASE>;
    close ENTRYBASE;

    if ($entrynumber !~ /^\d+$/ || $entrynumber > @child_parent) {
	print "invalid entry.\n" and die;
    }

    if (-e $basedir . "sections/$section/entries/$entrynumber.del" ||
	!-e $basedir . "sections/$section/entries/$entrynumber.hdl") {
	print "no such entry\n" and exit;
    }

    # prior to version 1.04, no "cnt"-file was written by default.
    my $numhits = 0;
    my $countfile = $basedir . "sections/$section/entries/$entrynumber.cnt";
    if (-e $countfile) {
	open my $cnt, "<$countfile";
	lock_sh $cnt;
	$numhits = <$cnt>;
	unlock $cnt;
	close $cnt;
	chomp $numhits;
    }
    $numhits++;

    # if two processes want to update $countfile simultaneously, at
    # worst we lose one hit. at least the count isn't completely
    # distorted (e.g., by interleaving writes)

    open my $cntw, ">$countfile";
    lock $cntw;
    print $cntw $numhits;
    unlock $cntw;
    close $cntw;

    load_entry $entrynumber, 1;

    print_html 'forum.hdr', 'navigation.tpl', 'fullentry.tpl';

    # we now wind up through @child_parent until we find a message
    # with "0" as parent and display the whole thread -- alternative:
    # print_entries $entrynumber;

    my $topparent = $entrynumber;
    my $nextparent = -1;
    my $index = $entrynumber - 1;
    do {
	my $entry = $child_parent[$index];
	chomp $entry;
	(my $child, $nextparent) = split /\|/, $entry;
	if ($nextparent > 0) { $topparent = $nextparent; }
	$index = $nextparent - 1;
    } while ($nextparent > 0);

    # we can cut off all entries up to, and not including,
    # this topmost parent (child of "0")
    @child_parent = @child_parent[++$index..$#child_parent];
    $entry_info{'topmost'} = $entrynumber;
    print_entries $topparent;


    # reload clicked entry for templates
    load_entry $entrynumber, 1;
    print_html 'addentry.tpl', 'forum.end';
}

sub num_lines
{
    my $file = shift;
    my $num = 0;
    if (-e $file) {
	open FH, "<$file";
	while (<FH>) { $num++; }
	close FH;
    }
    return $num;
}



sub write_entry
{
    (my $parent = $cgi->param("parent")) =~ /^\d+$/
	or print "Invalid parent.\n" and die;

    my ($email, $author, $password) =
	($cgi->param('email'), $cgi->param('author'), $cgi->param('password'));

    while ($author =~ /^REGISTERED/) { $author =~ s/^REGISTERED//; }

    my ($text, $headline) = ($cgi->param('text'), $cgi->param('headline'));

    my $entryinfo = '';

    my $authornum = -1;
    if (-e $basedir . "users/users.db") {
	open USERS, "<$basedir" . 'users/users.db';
	while (<USERS>) {
	    chomp;
	    my ($usernum, $name) = split /\|/;
	    if ($name eq $author) { $authornum = $usernum; last; }
	}
	close USERS;
    }

    if ($authornum >= 0) {
	open PW, "<$basedir" . "users/$authornum.pwl"
	    or print "cannot open password file $authornum.pwl" and die;
	my $stored = <PW>;
	close PW;

	verify_password $password, $stored
	    or print_html 'usr/wrongpw.html' and exit;

	open EMAIL, "<$basedir" . "users/$authornum.eml";
	my $registeredemail = <EMAIL> || "";
	close EMAIL;

	if (!$email || ($registeredemail eq $email)) { $email = "seeeml"; }
	$entryinfo = "REGISTERED$authornum\|" . localtime() ."\|$email";
    } else {
	$only_registered and print_html 'usr/notregistered.html' and exit;
	$entryinfo = "$author\|" . localtime() . "|$email";
    }

    my $entriesfile = $basedir . "sections/$section/entries/entries.db";

    if (-e $entriesfile && !-w $entriesfile) {
	print "$entriesfile must be WRITABLE for HTTP process.\n" and die;
    }

    if (!-w "$basedir" . "sections/$section/entries/") {
	print "Directory \"entries\" in section \"$section\" must be WRITABLE for HTTP process.\n" and die;
    }

    # we use cgiforum.conf for (exclusive) locking to ensure that the
    # new entry ID is unique

    open my $lockfile, "<cgiforum.conf";
    lock $lockfile;

    my $thisentry = (num_lines $entriesfile) + 1;
    $parent < $thisentry or print "invalid parent\n" and die;

    open INF,">$basedir"."sections\/$section\/entries\/$thisentry\.inf";
    print INF "$entryinfo\n" and close INF;

    open COUNT, ">$basedir" . "sections/$section/entries/$thisentry.cnt";
    print COUNT 0 and close COUNT;

    open HDL, ">$basedir" . "sections/$section/entries/$thisentry.hdl";
    print HDL $headline and close HDL;

    open MSG, ">$basedir" . "sections/$section/entries/$thisentry.msg";
    print MSG $text and close MSG;

    open ENTRIES, ">>$entriesfile";
    print ENTRIES "$thisentry\|$parent\n" and close ENTRIES;

    unlock $lockfile;
    close $lockfile;

    load_entry $thisentry, 1;
    print_html 'usr/entrysent.html';
}


sub print_section
{
    print_html 'forum.hdr';
    $entry_info{'topmost'} = 0;
    print_entries 0;
    print_html 'searchsection.tpl', 'newentry.tpl', 'forum.end';
}

sub search_section
{
    my $string = $cgi->param('searchstring');
    my $normal_template = load_template 'normalentry.tpl';
    my $path = $basedir . "sections/$section/entries";

    print_html 'forum.hdr';

    my @nums = ();

    foreach my $entry (<$path/*.msg>) {
	(my $num = $entry) =~ s/.*\/(.*)\.msg/$1/;
	unless (-e "$path/$num.del") { push @nums, $num; }
    }

    foreach my $num (sort {$b <=> $a} @nums) {
	my $hdl = file_to_scalar "$path/$num.hdl";
	if ($hdl =~ /$string/i) {
	    load_entry $num, $load_text;
	    print subst_variables $normal_template;
	} else {
	    my $msg = file_to_scalar "$path/$num.msg";
	    if ($msg =~ /$string/i) {
		load_entry $num, $load_text;
		print subst_variables $normal_template;
	    }
	}
    }

    print_html 'navigation.tpl', 'forum.end';
}

sub add_user
{
    $allow_registration or print_html 'usr/notallowed.html' and exit;

    (my $newuser = $cgi->param('name')) !~ /^REGISTERED/
	or print "Name can't start with REGISTERED.\n" and exit;
    my $newemail = $cgi->param('email');
    my ($newpw, $retypedpw) = ($cgi->param('newpw'), $cgi->param('retpw'));

    # uncomment this to prohibit multiple successive registrations
    # from one IP (you should impose this with a timeout):

    # my $lastip = "";
    # if (-e $basedir . 'users/lastip') {
    #    open LASTIP, "<$basedir" . 'users/lastip';
    #    $lastip = <LASTIP>;
    #    close LASTIP;
    #    chomp $lastip;
    # }
    # if ($lastip eq $ENV{'REMOTE_ADDR'}) {
    # 	 print_html 'usr/onepip.html' and exit;
    # }

    chomp $newuser;
    my $authornum = -1;
    my $lastusernum = -1;

    open my $lockfile, "<cgiforum.conf";
    lock $lockfile;

    if (-e $basedir . "users/users.db") {
	open USERS, "<$basedir" . "users/users.db";
	while (<USERS>) {
	    my ($usernum, $name) = split /\|/;
	    chomp $name;
	    if ($name eq $newuser) { $authornum = $usernum; }
	    $lastusernum = $usernum;
	}
	close USERS;
    }
    $authornum == -1 or print_html 'usr/notfree.html' and exit;
    $newpw eq $retypedpw or print_html 'usr/nopwmatch.html' and exit;

    my $newusernum = $lastusernum + 1;

    open EMAIL, ">>$basedir" . "users/$newusernum.eml";
    print EMAIL $newemail and close EMAIL;

    my $encpw = encrypt_password $newpw;
    my $pwfile = "$basedir" . "users/$newusernum.pwl";
    open PW, ">$pwfile";
    chmod 0600, $pwfile;
    print PW $encpw and close PW;

    open SYNTAX, ">$basedir" . "users/$newusernum.syn";
    print SYNTAX $newuser and close SYNTAX;

    # uncomment this to keep track of last IP
    # open LASTIP, ">$basedir" . "users/lastip";
    # print LASTIP $ENV{'REMOTE_ADDR'} and close LASTIP;

    open USERS, ">>$basedir" . "users/users.db";
    print USERS "$newusernum\|$newuser\n" and close USERS;

    unlock $lockfile;
    close $lockfile;

    print_html 'usr/registered.html';
}


print "Content-type: text/html\n\n";

# Load configuration file

open CONF, "<cgiforum.conf" or print "Cannot read cgiforum.conf.\n" and die;
while (<CONF>) {
    if (/^\#/) { next; }
    s/\r//g;
    chomp;
    my ($variable, $value) = split (/=/, $_);
    if ($variable =~ /basedir/i) {
	$basedir = $value;
	if (substr($basedir, -1, 1) ne '/') { $basedir .= '/'; }
    } elsif ($variable =~ /period_new/i) {
	$period_new = $value;
    } elsif ($variable =~ /default_day_limit/i) {
	$default_day_limit = $value;
    } elsif ($variable =~ /default_author/i) {
	$default_author = $value;
    } elsif ($variable =~ /default_headline/i) {
	$default_headline = $value;
    } elsif ($variable =~ /start_entry/i) {
	(my $sec = $variable) =~ s/start_entry_(.*)/$1/;
	$start_entries{$sec} = $value;
    } elsif ($variable =~ /load_text/i) {
	$load_text = $value;
    } elsif ($variable =~ /only_registered/i) {
	$only_registered = $value;
    } elsif ($variable =~ /allow_registration/i) {
	$allow_registration = $value;
    } elsif ($variable =~ /encryption_method/i) {
	$encryption_method = $value;
    }
}
close CONF;


# CGI parameter is called "thesection" since some browsers mistakenly
# treat "&section" as "&sect;ion"

$section = filter_input ($cgi->param('thesection') || 'default');

if (!-e "$basedir" . "sections/$section/html/wdays") {
    open WDAYS, "<$basedir" . "html/wdays";
} else {
    open WDAYS, "<$basedir" . "sections/$section/html/wdays";
}
while (<WDAYS>) {
    chomp;
    $expand_day{('Mon','Tue','Wed','Thu','Fri','Sat','Sun')[$. - 1]} = $_;
}

close WDAYS;


$entry_info{'basedir'} = $basedir;
$entry_info{'section'} = $section;

$entry_info{'end'} = $cgi->param('end') || 0;
$entry_info{'start'} = $cgi->param('start') || $default_day_limit;

if (!$cgi->param('action') || ($cgi->param('action') eq 'showsection')) {
    print_section;
} elsif ($cgi->param('action') eq 'addentry') {
    write_entry;
} elsif ($cgi->param('action') eq 'showentry') {
    print_entry filter_input($cgi->param('entry'));
} elsif ($cgi->param('action') eq 'adduser') {
    add_user;
} elsif ($cgi->param('action') eq 'search') {
    search_section;
}
